
# written to: 
# http://standards.freedesktop.org/shared-mime-info-spec/shared-mime-info-spec-0.13.html

import xdgbasedir
import os, re, sys, fnmatch

def mime():
    return [os.path.join(t, 'mime') for t in xdgbasedir.xdg_data_dirs()]

class GlobLookup(object):
    def __init__(self):
        self.literal = {}
        self.complex = {}
        self.extensions = {}
        mime_dirs = mime()
        for precedence, dir in enumerate(mime_dirs):
            mime_file = os.path.join(dir, 'globs')
            if os.access(mime_file, os.R_OK):
                self.update(precedence, mime_file)
        self.complex_by_size = self.complex.keys()
        self.complex_by_size.sort(lambda b, a: cmp(len(a), len(b)))

    def update(self, precedence, filename):
        literal = re.compile(r'^[^\[\*\?]+$')
        simple_glob = re.compile(r'^\*\.([^\[\*\?]+)$')
        for line in (t.rstrip('\n') for t in open(filename)):
            if line.startswith('#'):
                continue
            mimetype, glob = line.split(':')
            if literal.match(glob):
                self.add_literal(precedence, glob, mimetype)
                continue
            m = simple_glob.match(glob)
            if m:
                self.add_extension(precedence, m.groups()[0], mimetype)
            else:
                self.add_complex(precedence, glob, mimetype)
    
    def add_literal(self, precedence, glob, mimetype):
        self.literal.setdefault(glob, []).append((precedence, mimetype))
        
    def add_extension(self, precedence, extension, mimetype):
        self.extensions.setdefault(extension, []).append((precedence, mimetype))

    def add_complex(self, precedence, glob, mimetype):
        self.complex.setdefault(glob, []).append((precedence, mimetype))
    
    def __lookup(self, filename):
        # increasing; lower is higher priority
        precedence_sort = lambda a,b: cmp(a[0], b[0])
        def best_of(match_list):
            match_list.sort(precedence_sort)
            return match_list[0][1]
        if self.literal.has_key(filename):
            return best_of(self.literal[filename])
        for complex in self.complex_by_size:
            if fnmatch.fnmatch(filename, complex):
                return best_of(self.complex[complex])
        # find potential extensions in filename
        def extensions():
            ext = filename
            while ext:
                idx = ext.find('.')
                if idx == -1:
                    break
                ext = ext[idx+1:]
                yield ext
        for extension in extensions():
            if self.extensions.has_key(extension):
                return best_of(self.extensions[extension])
        return None
    
    def lookup(self, filename):
        rv = self.__lookup(filename)
        if rv: return rv
        filename = filename.lower()
        return self.__lookup(filename)

class MagicLookup(object):
    def __init__(self):
        # hashed by priority, then by mime type, then a list of headers
        # (multiple files could easily have conflicts otherwise)
        self.headers = {}
        mime_dirs = mime()
        for dir in mime_dirs:
            magic_file = os.path.join(dir, 'magic')
            if os.access(magic_file, os.R_OK):
                self.update(magic_file)

    def update(self, fname):
        fd = open(fname, 'rb')
        if fd.readline() != 'MIME-Magic\0\n':
            raise Exception("Not a Mime Magic file: %s" % fname)

        header_re = re.compile(r'^\[([0-9]+):([^\]]+)\]$')
        value_re = re.compile(r'^([0-9]*)>([0-9]+)=')
        options_re = re.compile(r'^(\~[0-9]+)?(\+[0-9]+)?$')

        # read a header, followed by a number of lines
        self.__buf = ''

        def skip_line():
            nl = self.__buf.find('\n')
            if nl == -1:
                self.__buf = ''
            else:
                self.__buf = self.__buf[nl+1:]
                
        def read_header():
            m = header_re.match(self.__buf)
            if not m:
                skip_line()
            else:
                self.__buf = self.__buf[m.end()+1:]
                priority, mime_type = m.groups()
                return int(priority), mime_type
                
        def read_line():
            # the next line will have indent, start_offset and the 
            # start of the value
            m = value_re.match(self.__buf)
            if not m:
                skip_line()
                return
            indent, start_offset = m.groups()
            try: indent = int(indent)
            except: indent = 0
            try: start_offset = int(start_offset)
            except: start_offset = 0
            self.__buf = self.__buf[m.end():]
            # the next two bytes are the length, in big-endian format
            length = (ord(self.__buf[0]) << 8) + ord(self.__buf[1])
            self.__buf = self.__buf[2:]
            # read the remainder of the value
            to_read = length - len(self.__buf)
            if to_read > 0:
                self.__buf += fd.read(to_read)
            value, self.__buf = self.__buf[:length], self.__buf[length:]
            # is the next thing a mask?
            if len(self.__buf) == 0:
                self.__buf += fd.read(1)
            if self.__buf[0] == '&':
                self.__buf = self.__buf[1:]
                to_read = length - len(self.__buf)
                if to_read > 0:
                    self.__buf += fd.read(to_read)
                mask, self.__buf = self.__buf[:length], self.__buf[length:]
            else:
                mask = '\xff' * length
            # anything remaining will end in a newline; so readline() 
            # does what we want. see whether or not we need to call it..
            if len(self.__buf) == 0 or self.__buf[-1] != '\n':
                self.__buf += fd.readline()
            word_size, range_length = 1, 1
            m = options_re.match(self.__buf)
            if m:
                for group in m.groups():
                    if not group:
                        continue
                    if group[0] == '~':
                        word_size = int(group[1:])
                    elif group[0] == '+':
                        range_length = int(group[1:])
                self.__buf = self.__buf[m.end()+1:]
            # fix the byte order, on little-endian systems
            if sys.byteorder == 'little' and word_size > 1:
                if len(value) % word_size != 0:
                    raise Exception("value is not an integer multiple of word size!")
                # make your sanity save now!
                fix = lambda x: ''.join([''.join(reversed(value[t:t+word_size])) for t in xrange(0,len(x)/word_size,word_size)])
                value = fix(value)
                mask = fix(mask)
            return { 'indent' : indent, 
                     'start_offset' : start_offset, 
                     'value' : value, 
                     'mask' : mask, 
                     'range_length' : range_length,
                     'word_size' : word_size }

        current_header = None
        while True:
            if self.__buf == '':
                nd = fd.readline()
                if not nd:
                    break
                self.__buf += nd
#            print "loop:", repr(self.__buf)
            if self.__buf[0] == '[':
                priority, mime_type = read_header()
                current_header = []
                self.headers.setdefault(priority, []).append((mime_type, current_header))
                line_stack = []
            else:
                if current_header == None:
                    raise Exception("non-header before header!")
                line = read_line()
                current_header.append(line)

    def lookup(self, data, priorities=None):
        def match_line(line):
            data_size = len(data)
            value_size = len(line['value'])

            def match_with(data_chunk):
                if line['mask'] != None:
                    masked = ''.join([chr(ord(t) & ord(line['mask'][i])) for i, t in enumerate(data_chunk)])
                else:
                    masked = data_chunk
                if masked == line['value']:
                    return True

            for i in range(line['range_length']):
                from_offset = line['start_offset'] + i
                to_offset = from_offset + value_size
                if to_offset > data_size:
                    continue
                if match_with(data[from_offset:to_offset]):
                    return True
            
        def match_lines(lines):
            # we need to maintain a current indent depth; we don't need to 
            # actually care what our parent was, as if we made it to checking 
            # we have necessarily succeeded
            depth = -1
            length = len(lines)
            for idx, line in enumerate(lines):
                indent = line['indent']
                if indent > depth+1:
                    continue
                if match_line(line):
                    if (idx+1 == length) or lines[idx+1]['indent'] <= indent:
                        # this is a match by itself
                        return True
                    depth = indent
                elif indent <= depth:
                    depth = indent - 1

        if not priorities:
            priorities = self.priorities()
        for priority in priorities:
            for mime_type, lines in self.headers[priority]:
                if match_lines(lines):
                    return mime_type
        return None

    def priorities(self):
        rv = self.headers.keys()
        rv.sort()
        rv.reverse()
        return rv

class LookupHelper:
    def __init__(self, remap_lookup=None):
        self.glob_lookup = GlobLookup()
        self.magic_lookup = MagicLookup()
        nontext_chars = "\x01\x02\x03\x04\x05\x06\x0e\x0f"\
                        "\x10\x11\x12\x13\x14\x15\x16\x17"\
                        "\x18\x19\x1a\x1c\x1d\x1e\x1f"
        self.nontext = {}
        for char in nontext_chars:
            self.nontext[char] = True
        self.remap_lookup = remap_lookup
            
    def is_binary(self, str):
        for char in str:
            if self.nontext.has_key(char):
                return True
        return False
        
    def __lookup(self, filename, data):
        # spec says we try >= 80 priority magic matchers, then filename, then the other matchers
        threshold = 80
        priorities = self.magic_lookup.priorities()
        rv = self.magic_lookup.lookup(data, [t for t in priorities if t >= threshold])
        if rv != None:
            return rv
        # then try guessing from filename
        rv = self.glob_lookup.lookup(filename)
        if rv != None:
            return rv
        # then try the other magic matchers
        rv = self.magic_lookup.lookup(data, [t for t in priorities if t < threshold])
        if rv != None:
            return rv
        # okay; fall-back behaviour is to return text/plain iff ! is_binary, 
        # otherwise application/octet-stream
        if self.is_binary(data):
            return 'application/octet-stream'
        else:
            return 'text/plain'
    
    def lookup(self, *args, **kwargs):
        rv = self.__lookup(*args, **kwargs)
        if self.remap_lookup and self.remap_lookup.has_key(rv):
            return self.remap_lookup[rv]
        else:
            return rv

if __name__ == '__main__':
    c = LookupHelper()
    print c.lookup('test.tar', '')

###
### vi:expandtab:sw=4:ts=4
###
